@@ -17,31 +17,41 @@
 import os
 import shutil
 import tarfile
+
 from contextlib import closing
 
-from cloudify_rest_client.exceptions import CloudifyClientError, \
-    UserUnauthorizedError
+from cloudify.utils import get_kerberos_indication
+from cloudify.cluster_status import CloudifyNodeType
+from cloudify_rest_client.exceptions import (CloudifyClientError,
+                                             UserUnauthorizedError)
 
 from . import init
 from .. import env
-from ..table import print_data
 from .. import utils
 from ..cli import cfy
 from .. import constants
 from ..cli import helptexts
+from ..env import get_rest_client
 from ..exceptions import CloudifyCliError
+from ..table import print_data, print_single
+from ..commands.cluster import _all_in_one_manager
+from ..commands.cluster import update_profile_logic as update_cluster_profile
+
 
 EXPORTED_KEYS_DIRNAME = '.exported-ssh-keys'
 EXPORTED_SSH_KEYS_DIR = os.path.join(env.PROFILES_DIR, EXPORTED_KEYS_DIRNAME)
 PROFILE_COLUMNS = ['name', 'manager_ip', 'manager_username', 'manager_tenant',
-                   'ssh_user', 'ssh_key_path', 'ssh_port',
+                   'ssh_user', 'ssh_key', 'ssh_port', 'kerberos_env',
                    'rest_port', 'rest_protocol', 'rest_certificate']
+CLUSTER_PROFILE_COLUMNS = PROFILE_COLUMNS[:1] + ['hostname', 'host_ip'] \
+    + PROFILE_COLUMNS[2:]
 
 
 @cfy.group(name='profiles')
-@cfy.options.verbose()
+@cfy.options.common_options
 def profiles():
-    """Handle Cloudify CLI profiles
+    """
+    Handle Cloudify CLI profiles
 
     Each profile can manage a single Cloudify manager.
 
@@ -54,12 +64,30 @@ def profiles():
         init.init_local_profile()
 
 
+def _format_cluster_profile(profile):
+    """
+    Format the list of cluster nodes for display in `cfy cluster show`,
+    we show the profile details of every stored cluster node.
+    """
+    common_attributes = {k: profile.get(k) for k in CLUSTER_PROFILE_COLUMNS}
+    nodes = []
+    for node in profile['cluster'][CloudifyNodeType.MANAGER]:
+        # merge the common attrs with node data, but rename node's name
+        # attribute to cluster_node, because the attribute 'name' is
+        # reserved for the profile name
+        node_data = dict(node)
+        node_data['hostname'] = node_data.pop('hostname')
+        nodes.append(dict(common_attributes, **node_data))
+    return nodes
+
+
 @profiles.command(name='show-current',
                   short_help='Retrieve current profile information')
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
 def show(logger):
-    """Shows your current active profile and it's properties
+    """
+    Shows your current active profile and it's properties
     """
     active_profile_name = env.get_active_profile()
     if active_profile_name == 'local':
@@ -68,15 +96,27 @@ def show(logger):
         return
 
     active_profile = _get_profile(env.get_active_profile())
-    print_data(PROFILE_COLUMNS, active_profile, 'Active profile:')
+    if active_profile.get('cluster'):
+
+        print_data(CLUSTER_PROFILE_COLUMNS,
+                   _format_cluster_profile(active_profile),
+                   'Cluster nodes in profile {0}:'
+                   .format(active_profile['name']),
+                   labels={
+                       'profile_name': 'Name',
+                       'hostname': 'Manager hostname',
+                       'host_ip': 'Manager ip'})
+    else:
+        print_single(PROFILE_COLUMNS, active_profile, 'Active profile:')
 
 
 @profiles.command(name='list',
                   short_help='List profiles')
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
-def list(logger):
-    """List all profiles
+def profiles_list(logger):
+    """
+    List all profiles
     """
     current_profile = env.get_active_profile()
 
@@ -108,12 +148,13 @@ def list(logger):
 @cfy.options.ssh_port
 @cfy.options.manager_username
 @cfy.options.manager_password
-@cfy.options.manager_tenant
+@cfy.options.manager_tenant()
 @cfy.options.rest_port
 @cfy.options.ssl_rest
 @cfy.options.rest_certificate
+@cfy.options.kerberos_env
 @cfy.options.skip_credentials_validation
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
 def use(manager_ip,
         profile_name,
@@ -143,18 +184,34 @@ def use(manager_ip,
             logger=logger,
             **kwargs)
     else:
+        kwargs.setdefault('manager_tenant', 'default_tenant')
         _create_profile(
             manager_ip=manager_ip,
             profile_name=profile_name,
             skip_credentials_validation=skip_credentials_validation,
             logger=logger,
             **kwargs)
+    if not env.profile.manager_username:
+        return
+    _update_cluster_profile_to_dict(logger)
+
+
+def _update_cluster_profile_to_dict(logger):
+    if isinstance(env.profile.cluster, list):   # noqa
+        env.profile.cluster = dict()
+        env.profile.save()
+    client = get_rest_client()
+    if not _all_in_one_manager(client):
+        update_cluster_profile(client, logger)
 
 
 def _switch_profile(manager_ip, profile_name, logger, **kwargs):
     # if using an existing profile, it is an error to provide any --option,
     # because the way to update an existing profile is `cfy profiles set`
+    if kwargs.get('manager_tenant') == env.profile.manager_tenant:
+        del kwargs['manager_tenant']
     provided_options = [key for key, value in kwargs.items() if value]
+
     if any(provided_options):
         logger.warning('Profile {0} already exists. '
                        'The passed in options are ignored: {1}. '
@@ -177,19 +234,19 @@ def _create_profile(
         rest_port,
         ssl,
         rest_certificate,
+        kerberos_env,
         skip_credentials_validation,
         logger):
     # If REST certificate is provided, then automatically
     # assume SSL.
     if rest_certificate:
         ssl = True
-
-    rest_protocol = constants.SECURED_REST_PROTOCOL if ssl else \
-        constants.DEFAULT_REST_PROTOCOL
-
+    rest_protocol, default_rest_port = _get_ssl_protocol_and_port(ssl)
     if not rest_port:
-        rest_port = constants.SECURED_REST_PORT if ssl else \
-            constants.DEFAULT_REST_PORT
+        rest_port = default_rest_port
+
+    # kerberos_env default is `False` and not `None`
+    kerberos_env = get_kerberos_indication(kerberos_env) or False
 
     logger.info('Attempting to connect to {0} through port {1}, using {2} '
                 '(SSL mode: {3})...'.format(manager_ip, rest_port,
@@ -206,15 +263,12 @@ def _create_profile(
         manager_username,
         manager_password,
         manager_tenant,
+        kerberos_env,
         skip_credentials_validation
     )
-
     init.init_manager_profile(profile_name=profile_name)
-    env.set_active_profile(profile_name)
-
     logger.info('Using manager {0} with port {1}'.format(
         manager_ip, rest_port))
-
     _set_profile_context(
         profile_name,
         provider_context,
@@ -227,14 +281,16 @@ def _create_profile(
         manager_tenant,
         rest_port,
         rest_protocol,
-        rest_certificate
+        rest_certificate,
+        kerberos_env
     )
+    env.set_active_profile(profile_name)
 
 
 @profiles.command(name='delete',
                   short_help='Delete a profile')
 @cfy.argument('profile-name')
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
 def delete(profile_name, logger):
     """Delete a profile
@@ -242,11 +298,11 @@ def delete(profile_name, logger):
     `PROFILE_NAME` is the IP of the manager the profile manages.
     """
     logger.info('Deleting profile {0}...'.format(profile_name))
-    try:
-        env.delete_profile(profile_name)
-        logger.info('Profile deleted')
-    except CloudifyCliError as ex:
-        logger.info(str(ex))
+    if not env.is_profile_exists(profile_name):
+        raise CloudifyCliError('Profile {0} does not exist'
+                               .format(profile_name))
+    env.delete_profile(profile_name)
+    logger.info('Profile deleted')
 
 
 def set_profile(profile_name,
@@ -258,23 +314,35 @@ def set_profile(profile_name,
                 ssh_port,
                 ssl,
                 rest_certificate,
+                rest_port,
+                kerberos_env,
                 skip_credentials_validation,
                 logger):
     """Set the profile name, manager username and/or password and/or tenant
     and/or ssl state (on/off) in the *current* profile
     """
     if not any([profile_name, ssh_user, ssh_key, ssh_port, manager_username,
-                manager_password, manager_tenant, ssl, rest_certificate]):
+                manager_password, manager_tenant, ssl is not None,
+                rest_certificate, kerberos_env is not None]):
         raise CloudifyCliError(
             "You must supply at least one of the following:  "
             "profile name, username, password, tenant, "
-            "ssl, rest certificate, ssh user, ssh key, ssh port")
+            "ssl, rest certificate, ssh user, ssh key, ssh port, kerberos env")
     username = manager_username or env.get_username()
     password = manager_password or env.get_password()
     tenant = manager_tenant or env.get_tenant_name()
+    protocol, port = _get_ssl_protocol_and_port(ssl)
+    if rest_port is not None:
+        port = rest_port
 
     if not skip_credentials_validation:
-        _validate_credentials(username, password, tenant, rest_certificate)
+        _validate_credentials(username,
+                              password,
+                              tenant,
+                              rest_certificate,
+                              protocol,
+                              port,
+                              kerberos_env)
     old_name = None
     if profile_name:
         if profile_name == 'local':
@@ -293,12 +361,13 @@ def set_profile(profile_name,
     if manager_tenant:
         logger.info('Setting tenant to `{0}`'.format(manager_tenant))
         env.profile.manager_tenant = manager_tenant
-    if ssl is not None:
-        _set_profile_ssl(ssl, logger)
     if rest_certificate:
         logger.info(
             'Setting rest certificate to `{0}`'.format(rest_certificate))
         env.profile.rest_certificate = rest_certificate
+    if rest_port:
+        logger.info('Setting rest port to `{0}'.format(rest_port))
+        env.profile.rest_port = rest_port
     if ssh_user:
         logger.info('Setting ssh user to `{0}`'.format(ssh_user))
         env.profile.ssh_user = ssh_user
@@ -308,6 +377,11 @@ def set_profile(profile_name,
     if ssh_port:
         logger.info('Setting ssh port to `{0}`'.format(ssh_port))
         env.profile.ssh_port = ssh_port
+    if kerberos_env is not None:
+        logger.info('Setting kerberos_env to `{0}`'.format(kerberos_env))
+        env.profile.kerberos_env = kerberos_env
+    if ssl is not None:
+        _set_profile_ssl(ssl, rest_port, logger)
 
     env.profile.save()
     if old_name is not None:
@@ -316,29 +390,36 @@ def set_profile(profile_name,
     logger.info('Settings saved successfully')
 
 
-def _set_profile_ssl(ssl, logger):
-    ssl = str(ssl).lower()
-    if ssl == 'on':
+def _set_profile_ssl(ssl, rest_port, logger):
+    if ssl is None:
+        raise CloudifyCliError('Internal error: SSL must be either `on` or '
+                               '`off`')
+
+    protocol, port = _get_ssl_protocol_and_port(ssl)
+    if rest_port is not None:
+        port = rest_port
+    if protocol == constants.SECURED_REST_PROTOCOL:
         logger.info('Enabling SSL in the local profile')
-        port = constants.SECURED_REST_PORT
-        protocol = constants.SECURED_REST_PROTOCOL
-    elif ssl == 'off':
-        logger.info('Disabling SSL in the local profile')
-        port = constants.DEFAULT_REST_PORT
-        protocol = constants.DEFAULT_REST_PROTOCOL
     else:
-        raise CloudifyCliError('SSL must be either `on` or `off`')
+        logger.info('Disabling SSL in the local profile')
 
     env.profile.rest_port = port
     env.profile.rest_protocol = protocol
 
-    if env.profile.cluster:
-        for node in env.profile.cluster:
+    manager_cluster = env.profile.cluster.get(CloudifyNodeType.MANAGER)
+    if manager_cluster:
+        missing_certs = []
+        for node in manager_cluster:
             node['rest_port'] = port
             node['rest_protocol'] = protocol
-            node['trust_all'] = True
-            logger.info('Enabling SSL for {0} without certificate validation'
-                        .format(node['manager_ip']))
+            logger.info('Enabling SSL for {0}'.format(node['host_ip']))
+            if not node.get('cert'):
+                missing_certs.append(node['hostname'])
+        if missing_certs:
+            logger.warning('The following cluster nodes have no certificate '
+                           'set: {0}'.format(', '.join(missing_certs)))
+            logger.warning('If required, set the certificates for those '
+                           'nodes using `cfy profiles set-cluster`')
 
 
 @profiles.command(
@@ -347,14 +428,16 @@ def _set_profile_ssl(ssl, logger):
 @cfy.options.profile_name
 @cfy.options.manager_username
 @cfy.options.manager_password
-@cfy.options.manager_tenant
+@cfy.options.manager_tenant()
 @cfy.options.ssh_user
 @cfy.options.ssh_key
 @cfy.options.ssh_port
 @cfy.options.ssl_state
 @cfy.options.rest_certificate
+@cfy.options.rest_port
+@cfy.options.kerberos_env
 @cfy.options.skip_credentials_validation
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
 def set_cmd(profile_name,
             manager_username,
@@ -365,6 +448,8 @@ def set_cmd(profile_name,
             ssh_port,
             ssl,
             rest_certificate,
+            rest_port,
+            kerberos_env,
             skip_credentials_validation,
             logger):
     return set_profile(profile_name,
@@ -374,8 +459,10 @@ def set_cmd(profile_name,
                        ssh_user,
                        ssh_key,
                        ssh_port,
-                       ssl,
+                       _get_ssl_indication(ssl),
                        rest_certificate,
+                       rest_port,
+                       get_kerberos_indication(kerberos_env),
                        skip_credentials_validation,
                        logger)
 
@@ -395,11 +482,12 @@ def set_cluster(cluster_node_name,
                 ssh_port,
                 rest_certificate,
                 logger):
-    """Set connection options for a cluster node.
+    """Set connection options for a Manager cluster node.
 
-    `CLUSTER_NODE_NAME` is the name of the cluster node to set options for.
+    `CLUSTER_NODE_NAME` is the Manager cluster node name to set options for.
     """
-    if not env.profile.cluster:
+    manager_cluster = env.profile.cluster.get(CloudifyNodeType.MANAGER)
+    if not manager_cluster:
         err = CloudifyCliError('The current profile is not a cluster profile!')
         err.possible_solutions = [
             "Select a different profile using `cfy profiles use`",
@@ -408,8 +496,8 @@ def set_cluster(cluster_node_name,
         raise err
 
     changed_node = None
-    for node in env.profile.cluster:
-        if node['name'] == cluster_node_name:
+    for node in manager_cluster:
+        if node['hostname'] == cluster_node_name:
             changed_node = node
             break
     else:
@@ -420,13 +508,18 @@ def set_cluster(cluster_node_name,
         (ssh_user, 'ssh_user', 'ssh user'),
         (ssh_key, 'ssh_key', 'ssh key'),
         (ssh_port, 'ssh_port', 'ssh port'),
-        (rest_certificate, 'cert', 'rest certificate'),
     ]:
         if source:
             changed_node[target] = source
             logger.info('Node {0}: setting {1} to `{2}`'
                         .format(cluster_node_name, label, source))
-
+    if rest_certificate:
+        changed_node['cert'] = rest_certificate
+        changed_node['trust_all'] = False
+        changed_node['rest_protocol'] = 'https'
+        logger.info('Node {0}: setting rest-certificate to `{1}` and enabling '
+                    'certificate verification'
+                    .format(cluster_node_name, source))
     env.profile.save()
     logger.info('Settings saved successfully')
 
@@ -440,8 +533,9 @@ def set_cluster(cluster_node_name,
 @cfy.options.ssh_user_flag
 @cfy.options.ssh_key_flag
 @cfy.options.rest_certificate_flag
+@cfy.options.kerberos_env_flag
 @cfy.options.skip_credentials_validation
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
 def unset(manager_username,
           manager_password,
@@ -449,15 +543,16 @@ def unset(manager_username,
           ssh_user,
           ssh_key,
           rest_certificate,
+          kerberos_env,
           skip_credentials_validation,
           logger):
     """Clear the manager username and/or password and/or tenant
     from the *current* profile
     """
     if not any([manager_username, manager_password, manager_tenant,
-                rest_certificate, ssh_user, ssh_key]):
+                rest_certificate, ssh_user, ssh_key, kerberos_env]):
         raise CloudifyCliError("You must choose at least one of the following:"
-                               " username, password, tenant, "
+                               " username, password, tenant, kerberos_env, "
                                "rest certificate, ssh user, ssh key")
     if manager_username:
         username = os.environ.get(constants.CLOUDIFY_USERNAME_ENV)
@@ -473,12 +568,18 @@ def unset(manager_username,
         tenant = env.profile.manager_tenant
     if rest_certificate:
         cert = os.environ.get(constants.LOCAL_REST_CERT_FILE) \
-               or env.get_default_rest_cert_local_path()
+            or env.get_default_rest_cert_local_path()
     else:
         cert = None
 
     if not skip_credentials_validation:
-        _validate_credentials(username, password, tenant, cert)
+        _validate_credentials(username,
+                              password,
+                              tenant,
+                              cert,
+                              env.profile.rest_protocol,
+                              env.profile.rest_port,
+                              None)
 
     if manager_username:
         logger.info('Clearing manager username')
@@ -498,6 +599,9 @@ def unset(manager_username,
     if ssh_key:
         logger.info('Clearing ssh key')
         env.profile.ssh_key = None
+    if kerberos_env:
+        logger.info('Clearing kerberos_env')
+        env.profile.kerberos_env = None
     env.profile.save()
     logger.info('Settings saved successfully')
 
@@ -506,7 +610,7 @@ def unset(manager_username,
                   short_help='Export all profiles to an archive')
 @cfy.options.include_keys(helptexts.EXPORT_SSH_KEYS)
 @cfy.options.optional_output_path
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
 def export_profiles(include_keys, output_path, logger):
     """Export all profiles to a file
@@ -541,7 +645,7 @@ def export_profiles(include_keys, output_path, logger):
                   short_help='Import profiles from an archive')
 @cfy.argument('archive-path')
 @cfy.options.include_keys(helptexts.IMPORT_SSH_KEYS)
-@cfy.options.verbose()
+@cfy.options.common_options
 @cfy.pass_logger
 def import_profiles(archive_path, include_keys, logger):
     """Import profiles from a profiles archive
@@ -635,7 +739,7 @@ def _get_profile(profile_name):
 def _assert_manager_available(client, profile_name):
     try:
         return client.manager.get_status()
-    except UserUnauthorizedError, e:
+    except UserUnauthorizedError as e:
         raise CloudifyCliError(
             "Can't use manager {0}\n{1}.".format(
                 profile_name,
@@ -648,7 +752,7 @@ def _assert_manager_available(client, profile_name):
     # with situations like No route to host and the likes.
     except Exception as ex:
         raise CloudifyCliError(
-            "Can't use manager {0}. {1}".format(profile_name, str(ex.message)))
+            "Can't use manager {0}. {1}".format(profile_name, ex))
 
 
 def _get_provider_context(profile_name,
@@ -659,6 +763,7 @@ def _get_provider_context(profile_name,
                           manager_username,
                           manager_password,
                           manager_tenant,
+                          kerberos_env,
                           skip_credentials_validation):
     try:
         client = _get_client_and_assert_manager(
@@ -669,7 +774,8 @@ def _get_provider_context(profile_name,
             rest_certificate,
             manager_username,
             manager_password,
-            manager_tenant
+            manager_tenant,
+            kerberos_env
         )
     except CloudifyCliError:
         if skip_credentials_validation:
@@ -690,7 +796,8 @@ def _get_client_and_assert_manager(profile_name,
                                    rest_certificate=None,
                                    manager_username=None,
                                    manager_password=None,
-                                   manager_tenant=None):
+                                   manager_tenant=None,
+                                   kerberos_env=None):
     # Attempt to update the profile with an existing profile context, if one
     # is available. This is relevant in case the user didn't pass a username
     # or a password, and was expecting them to be taken from the old profile
@@ -701,10 +808,10 @@ def _get_client_and_assert_manager(profile_name,
         rest_port=rest_port,
         rest_protocol=rest_protocol,
         rest_cert=rest_certificate,
-        skip_version_check=True,
         username=manager_username,
         password=manager_password,
-        tenant_name=manager_tenant
+        tenant_name=manager_tenant,
+        kerberos_env=kerberos_env
     )
 
     _assert_manager_available(client, profile_name)
@@ -722,7 +829,8 @@ def _set_profile_context(profile_name,
                          manager_tenant,
                          rest_port,
                          rest_protocol,
-                         rest_certificate):
+                         rest_certificate,
+                         kerberos_env):
 
     profile = env.get_profile_context(profile_name)
     profile.provider_context = provider_context
@@ -745,6 +853,7 @@ def _set_profile_context(profile_name,
     profile.ssh_port = ssh_port or constants.REMOTE_EXECUTION_PORT
     profile.rest_protocol = rest_protocol
     profile.rest_certificate = rest_certificate
+    profile.kerberos_env = kerberos_env
 
     profile.save()
 
@@ -763,14 +872,34 @@ def _is_manager_secured(response_history):
     return False
 
 
+def _get_ssl_indication(ssl):
+    if ssl is None:
+        return None
+    return str(ssl).lower() == 'on'
+
+
+def _get_ssl_protocol_and_port(ssl):
+    if ssl is not None:
+        protocol, port = (constants.SECURED_REST_PROTOCOL,
+                          constants.SECURED_REST_PORT) if ssl else \
+            (constants.DEFAULT_REST_PROTOCOL, constants.DEFAULT_REST_PORT)
+    else:
+        protocol, port = None, None
+    return protocol, port
+
+
 @cfy.pass_logger
-def _validate_credentials(username, password, tenant, certificate, logger):
+def _validate_credentials(username, password, tenant, certificate, protocol,
+                          rest_port, kerberos_env, logger):
     logger.info('Validating credentials...')
     _get_client_and_assert_manager(
         profile_name=env.profile.profile_name,
         manager_username=username,
         manager_password=password,
         manager_tenant=tenant,
-        rest_certificate=certificate
+        rest_certificate=certificate,
+        rest_protocol=protocol,
+        rest_port=rest_port,
+        kerberos_env=kerberos_env
     )
     logger.info('Credentials validated')
