@@ -17,25 +17,34 @@
 
 import os
 import json
+import errno
 import types
 import shutil
-import pkgutil
 import getpass
 import tempfile
 import itertools
-from base64 import urlsafe_b64encode
+from base64 import b64encode
+from contextlib import contextmanager
 
 import yaml
 import requests
 
 from cloudify_rest_client import CloudifyClient
 from cloudify_rest_client.client import HTTPClient
-from cloudify_rest_client.exceptions import (CloudifyClientError,
-                                             RemovedFromCluster,
-                                             NotClusterMaster)
+from cloudify.cluster_status import CloudifyNodeType
+from cloudify_rest_client.utils import is_kerberos_env
+from cloudify_rest_client.exceptions import CloudifyClientError
+
 from . import constants
 from .exceptions import CloudifyCliError
 
+try:
+    from fabric import Connection
+    from paramiko import AuthenticationException
+except ImportError:
+    Connection = None
+
+
 _ENV_NAME = 'manager'
 DEFAULT_LOG_FILE = os.path.expanduser(
     '{0}/cloudify-{1}/cloudify-cli.log'.format(
@@ -45,26 +54,23 @@ CLOUDIFY_WORKDIR = os.path.join(
     os.environ.get('CFY_WORKDIR', os.path.expanduser('~')),
     constants.CLOUDIFY_BASE_DIRECTORY_NAME)
 PROFILES_DIR = os.path.join(CLOUDIFY_WORKDIR, 'profiles')
-ACTIVE_PRO_FILE = os.path.join(CLOUDIFY_WORKDIR, 'active.profile')
-MULTIPLE_LOCAL_BLUEPRINTS = (
-    os.environ.get('CFY_MULTIPLE_BLUEPRINTS', 'true') == 'true')
+ACTIVE_PROFILE = os.path.join(CLOUDIFY_WORKDIR, 'active.profile')
 CLUSTER_RETRY_INTERVAL = 5
 
 
 def delete_profile(profile_name):
     if is_profile_exists(profile_name):
         profile_dir = get_profile_dir(profile_name)
-        shutil.rmtree(profile_dir)
-    else:
-        raise CloudifyCliError(
-            'Profile {0} does not exist'.format(profile_name))
+        if profile_dir:
+            shutil.rmtree(profile_dir)
 
 
 def is_profile_exists(profile_name):
-    try:
-        return os.path.isfile(get_context_path(profile_name))
-    except CloudifyCliError:
+    base_dir = get_profile_dir(profile_name)
+    if not base_dir:
         return False
+    return (os.path.isfile(os.path.join(base_dir, 'context.json')) or
+            os.path.isfile(os.path.join(base_dir, 'context')))
 
 
 def assert_profile_exists(profile_name):
@@ -76,20 +82,30 @@ def assert_profile_exists(profile_name):
 
 def set_active_profile(profile_name):
     global profile
-    with open(ACTIVE_PRO_FILE, 'w+') as active_profile:
+    with open(ACTIVE_PROFILE, 'w+') as active_profile:
         active_profile.write(profile_name)
     profile = get_profile_context(profile_name, suppress_error=True)
 
 
 def get_active_profile():
-    if os.path.isfile(ACTIVE_PRO_FILE):
-        with open(ACTIVE_PRO_FILE) as active_profile:
+    try:
+        with open(ACTIVE_PROFILE) as active_profile:
             return active_profile.read().strip()
-    else:
-        # We return None explicitly as no profile is active.
+    except IOError as e:
+        if e.errno != errno.ENOENT:
+            raise
         return None
 
 
+def set_target_manager(manager_host):
+    global target_manager
+    target_manager = manager_host
+
+
+def get_target_manager():
+    return target_manager
+
+
 def get_profile_names():
     # TODO: This is too.. ambiguous. We should change it so there are
     # no exclusions.
@@ -104,7 +120,8 @@ def assert_manager_active():
     if not is_manager_active():
         raise CloudifyCliError(
             'This command is only available when using a manager. '
-            'You need to run run `cfy profiles use MANAGER_IP`')
+            'Please use the `cfy profiles use` command to connect '
+            'to a Cloudify Manager.')
 
 
 def assert_local_active():
@@ -119,14 +136,15 @@ def assert_credentials_set():
                 'You can set it in the profile by running ' \
                 '`cfy profiles set {1}`, or you can set the `CLOUDIFY_{2}` ' \
                 'environment variable.'
-    if not get_username():
-        raise CloudifyCliError(
-            error_msg.format('Username', '--manager-username', 'USERNAME')
-        )
-    if not get_password():
-        raise CloudifyCliError(
-            error_msg.format('Password', '--manager-password', 'PASSWORD')
-        )
+    if not get_kerberos_env():
+        if not get_username():
+            raise CloudifyCliError(
+                error_msg.format('Username', '--manager-username', 'USERNAME')
+            )
+        if not get_password():
+            raise CloudifyCliError(
+                error_msg.format('Password', '--manager-password', 'PASSWORD')
+            )
     if not get_tenant_name():
         raise CloudifyCliError(
             error_msg.format('Tenant', '--manager-tenant', 'TENANT')
@@ -148,23 +166,85 @@ def is_manager_active():
 
 
 def get_profile_context(profile_name=None, suppress_error=False):
-    profile_name = profile_name or get_active_profile()
     if profile_name == 'local':
+        return ProfileContext({}, profile_name='local')
+
+    profile_name = profile_name or get_active_profile()
+    loaded = None
+    path = get_context_path(profile_name)
+    if path:
+        try:
+            with open(path) as f:
+                loaded = json.load(f)
+        except IOError as e:
+            if e.errno != errno.ENOENT:
+                raise
+
+    if not loaded:
+        loaded = _try_load_yaml_profile(profile_name)
+
+    if not loaded:
         if suppress_error:
-            return ProfileContext()
-        raise CloudifyCliError('Local profile does not have context')
+            return ProfileContext({})
+        raise CloudifyCliError('No context for profile {0}'
+                               .format(profile_name))
+    return ProfileContext(loaded, profile_name)
+
+
+class _ProfileLoader(yaml.SafeLoader):
+    """A yaml Loader that can load Cloudify 5.1 profiles
+
+    It supports python/unicode, which was commonly present in py2-created
+    profiles.
+    """
+
+
+def _load_str(loader, node):
+    return node.value
+
+
+_ProfileLoader.add_constructor(u'tag:yaml.org,2002:python/unicode', _load_str)
+
+
+def _try_load_yaml_profile(profile_name):
+    """Try to load the profile from the yaml context file.
+
+    This keeps compatibility with Cloudify 5.1 and earlier, who stored
+    the context in a yaml file. We will still load them, but we won't
+    store them anymore.
+    """
+    base_dir = get_profile_dir(profile_name)
+    if not base_dir:
+        return
     try:
-        path = get_context_path(profile_name)
-        with open(path) as f:
-            return yaml.load(f.read())
-    except CloudifyCliError:
-        if suppress_error:
-            return ProfileContext()
-        raise
+        with open(os.path.join(base_dir, 'context')) as f:
+            # dropping the object tag from yaml, so that we load the
+            # yaml as just a dict and not as an object
+            data = f.read().replace('!CloudifyProfileContext', '')
+            context = yaml.load(data, Loader=_ProfileLoader)
+    except IOError as e:
+        if e.errno != errno.ENOENT:
+            raise
+        return
+    return context
+
+
+def config_initialized_with_logging():
+    """
+    This is for the Windows agent: plugin URLs from
+    import_resolver are written to config.yaml during installation, so we can
+    have a scenario where config exists but has no logger paths defined.
+    """
+    has_logging = False
+    if os.path.isfile(os.path.join(CLOUDIFY_WORKDIR, 'config.yaml')):
+        with open(os.path.join(CLOUDIFY_WORKDIR, 'config.yaml'), 'r') as f:
+            has_logging = ('logging' in f.read())
+    return has_logging
 
 
 def is_initialized(profile_name=None):
-    """Check if a profile or an environment is initialized.
+    """
+    Check if a profile or an environment is initialized.
 
     If profile_name is provided, it will check if the profile
     is initialized. If not, it will just check that workenv is.
@@ -172,28 +252,21 @@ def is_initialized(profile_name=None):
     if profile_name:
         return get_profile_dir(profile_name) is not None
     else:
-        return os.path.isfile(os.path.join(CLOUDIFY_WORKDIR, 'config.yaml'))
+        return config_initialized_with_logging()
 
 
-def get_context_path(profile_name, suppress_error=False):
-    base_dir = get_profile_dir(profile_name, suppress_error)
+def get_context_path(profile_name):
+    base_dir = get_profile_dir(profile_name)
     if not base_dir:
         return
-    return os.path.join(
-        base_dir,
-        constants.CLOUDIFY_PROFILE_CONTEXT_FILE_NAME
-    )
+    return os.path.join(base_dir, 'context.json')
 
 
-def get_profile_dir(profile_name=None, suppress_error=False):
+def get_profile_dir(profile_name=None):
     active_profile = profile_name or get_active_profile()
     if active_profile and os.path.isdir(
             os.path.join(PROFILES_DIR, active_profile)):
         return os.path.join(PROFILES_DIR, active_profile)
-    elif suppress_error:
-        return
-    else:
-        raise CloudifyCliError('Profile directory does not exist')
 
 
 def raise_uninitialized():
@@ -205,6 +278,13 @@ def raise_uninitialized():
     raise error
 
 
+def is_cluster(client_profile=None):
+    if client_profile is None:
+        client_profile = profile
+    return (not isinstance(client_profile.cluster, list) and
+            client_profile.cluster.get(CloudifyNodeType.MANAGER))
+
+
 def get_rest_client(client_profile=None,
                     rest_host=None,
                     rest_port=None,
@@ -214,8 +294,8 @@ def get_rest_client(client_profile=None,
                     password=None,
                     tenant_name=None,
                     trust_all=False,
-                    skip_version_check=False,
-                    cluster=None):
+                    cluster=None,
+                    kerberos_env=None):
     if client_profile is None:
         client_profile = profile
     rest_host = rest_host or client_profile.manager_ip
@@ -228,70 +308,53 @@ def get_rest_client(client_profile=None,
     trust_all = trust_all or get_ssl_trust_all()
     headers = get_auth_header(username, password)
     headers[constants.CLOUDIFY_TENANT_HEADER] = tenant_name
-    cluster = cluster or client_profile.cluster
+    cluster = cluster or is_cluster(client_profile)
+    kerberos_env = kerberos_env \
+        if kerberos_env is not None else client_profile.kerberos_env
 
-    if not username:
-        raise CloudifyCliError('Command failed: Missing Username')
-
-    if not password:
-        raise CloudifyCliError('Command failed: Missing password')
-
-    if rest_protocol == constants.SECURED_REST_PROTOCOL and \
-            (not rest_cert or not os.path.isfile(rest_cert)):
-        raise CloudifyCliError('Command failed: Trying to work with https '
-                               'protocol without certificate. Make sure '
-                               '`rest_certificate` is defined in your profile '
-                               'to be the path to the valid rest public '
-                               'certificate file')
+    if kerberos_env is False \
+            or (kerberos_env is None and not is_kerberos_env()):
+        if not username:
+            raise CloudifyCliError('Command failed: Missing Username')
+        if not password:
+            raise CloudifyCliError('Command failed: Missing password')
 
     if cluster:
-        client = CloudifyClusterClient(
-            host=rest_host,
-            port=rest_port,
-            protocol=rest_protocol,
-            headers=headers,
-            cert=rest_cert,
-            trust_all=trust_all,
-            profile=client_profile)
-
+        client = CloudifyClusterClient(host=rest_host,
+                                       port=rest_port,
+                                       protocol=rest_protocol,
+                                       headers=headers,
+                                       cert=rest_cert,
+                                       trust_all=trust_all,
+                                       profile=client_profile,
+                                       kerberos_env=kerberos_env)
     else:
-        client = CloudifyClient(
-            host=rest_host,
-            port=rest_port,
-            protocol=rest_protocol,
-            headers=headers,
-            cert=rest_cert,
-            trust_all=trust_all)
-
-    # TODO: Put back version check after we've solved the problem where
-    # a new CLI is used with an older manager on `cfy upgrade`.
-    if skip_version_check or True:
-        return client
-
-    cli_version, manager_version = get_cli_manager_versions(client)
-
-    if cli_version == manager_version:
-        return client
-    elif not manager_version:
-        return client
-    else:
-        raise CloudifyCliError(
-            'CLI and manager versions do not match\n'
-            'CLI Version: {0}\n'
-            'Manager Version: {1}'.format(cli_version, manager_version))
+        client = CloudifyClient(host=rest_host,
+                                port=rest_port,
+                                protocol=rest_protocol,
+                                headers=headers,
+                                cert=rest_cert,
+                                trust_all=trust_all,
+                                kerberos_env=kerberos_env)
+    return client
 
 
 def build_manager_host_string(ssh_user='', ip=''):
+    ip = ip or profile.manager_ip
+    return build_host_string(ip, ssh_user)
+
+
+def build_host_string(ip, ssh_user=''):
     ssh_user = ssh_user or profile.ssh_user
     if not ssh_user:
-        raise CloudifyCliError('Manager `ssh_user` is not set '
-                               'in Cloudify CLI settings')
-    ip = ip or profile.manager_ip
+        raise CloudifyCliError('`ssh_user` is not set in the current '
+                               'profile. Please run '
+                               '`cfy profiles set --ssh-user <ssh-user>`.')
     return '{0}@{1}'.format(ssh_user, ip)
 
 
 def get_default_rest_cert_local_path():
-    base_dir = get_profile_dir(suppress_error=True) or CLOUDIFY_WORKDIR
+    base_dir = get_profile_dir() or CLOUDIFY_WORKDIR
     return os.path.join(base_dir, constants.PUBLIC_REST_CERT)
 
 
@@ -334,6 +397,12 @@ def get_tenant_name(from_profile=None):
     return tenant or from_profile.manager_tenant
 
 
+def get_kerberos_env(from_profile=None):
+    if from_profile is None:
+        from_profile = profile
+    return from_profile.kerberos_env
+
+
 def get_ssl_cert(from_profile=None):
     """Return the path to a local copy of the manager's public certificate.
 
@@ -366,17 +435,12 @@ def get_ssl_trust_all():
     return False
 
 
-def get_version_data():
-    data = pkgutil.get_data('cloudify_cli', 'VERSION')
-    return json.loads(data)
-
-
 def get_manager_version_data(rest_client=None):
     if not rest_client:
         if not get_profile_context(suppress_error=True):
             return None
         try:
-            rest_client = get_rest_client(skip_version_check=True)
+            rest_client = get_rest_client()
         except CloudifyCliError:
             return None
     try:
@@ -387,93 +451,48 @@ def get_manager_version_data(rest_client=None):
     return version_data
 
 
-def get_cli_manager_versions(rest_client):
-    manager_version_data = get_manager_version_data(rest_client)
-    cli_version = get_version_data().get('version')
+class ProfileContext(object):
+    def __init__(self, context=None, profile_name=None):
+        self._context = {
+            'name': profile_name,
+            'manager_ip': None,
+            'ssh_key': None,
+            'ssh_port': 22,
+            'ssh_user': None,
+            'provider_context': {},
+            'manager_username': None,
+            'manager_password': None,
+            'manager_tenant': None,
+            'rest_port': constants.DEFAULT_REST_PORT,
+            'rest_protocol': constants.DEFAULT_REST_PROTOCOL,
+            'rest_certificate': None,
+            'kerberos_env': False,
+            'cluster': {}
+        }
+        if context:
+            self._context.update(context)
 
-    if not manager_version_data:
-        return cli_version, None
-    else:
-        manager_version = manager_version_data.get('version')
-        return cli_version, manager_version
+    def __getattr__(self, name):
+        return self._context[name]
 
-
-class ProfileContext(yaml.YAMLObject):
-    yaml_tag = u'!CloudifyProfileContext'
-    yaml_loader = yaml.Loader
-
-    def __init__(self, profile_name=None):
-        # Note that __init__ is not called when loading from yaml.
-        # When adding a new ProfileContext attribute, make sure that
-        # all methods handle the case when the attribute is missing
-        self._profile_name = profile_name
-        self.manager_ip = None
-        self.ssh_key = None
-        self._ssh_port = None
-        self.ssh_user = None
-        self.provider_context = dict()
-        self.manager_username = None
-        self.manager_password = None
-        self.manager_tenant = None
-        self.rest_port = constants.DEFAULT_REST_PORT
-        self.rest_protocol = constants.DEFAULT_REST_PROTOCOL
-        self.rest_certificate = None
-        self._cluster = []
+    def __setattr__(self, name, value):
+        if name in ['_context', 'profile_name']:
+            super(ProfileContext, self).__setattr__(name, value)
+        else:
+            self._context[name] = value
 
     def to_dict(self):
-        return dict(
-            name=self.profile_name,
-            manager_ip=self.manager_ip,
-            ssh_key_path=self.ssh_key,
-            ssh_port=self.ssh_port,
-            ssh_user=self.ssh_user,
-            provider_context=self.provider_context,
-            manager_username=self.manager_username,
-            manager_tenant=self.manager_tenant,
-            rest_port=self.rest_port,
-            rest_protocol=self.rest_protocol,
-            rest_certificate=self.rest_certificate,
-            cluster=self.cluster
-        )
-
-    @property
-    def ssh_port(self):
-        return self._ssh_port
-
-    @ssh_port.setter
-    def ssh_port(self, ssh_port):
-        # If the port is int, we want to change it to a string. Otherwise,
-        # leave None as is
-        ssh_port = str(ssh_port) if ssh_port else None
-        self._ssh_port = ssh_port
+        ctx = self._context.copy()
+        ctx['name'] = self.profile_name
+        return ctx
 
     @property
     def profile_name(self):
-        return getattr(self, '_profile_name', None) \
-            or getattr(self, 'manager_ip', None)
+        return self._context['name'] or self._context['manager_ip']
 
     @profile_name.setter
-    def profile_name(self, profile_name):
-        self._profile_name = profile_name
-
-    @property
-    def cluster(self):
-        # default the ._cluster attribute here, so that all callers can use it
-        # as just ._cluster, even if it's not present in the source yaml
-        if not hasattr(self, '_cluster'):
-            self._cluster = []
-        return self._cluster
-
-    @cluster.setter
-    def cluster(self, cluster):
-        self._cluster = cluster
-
-    def _get_context_path(self):
-        init_path = get_profile_dir(self.profile_name)
-        context_path = os.path.join(
-            init_path,
-            constants.CLOUDIFY_PROFILE_CONTEXT_FILE_NAME)
-        return context_path
+    def profile_name(self, value):
+        self._context['name'] = value
 
     @property
     def workdir(self):
@@ -489,18 +508,19 @@ class ProfileContext(yaml.YAMLObject):
             os.makedirs(workdir)
         target_file_path = os.path.join(
             workdir,
-            constants.CLOUDIFY_PROFILE_CONTEXT_FILE_NAME)
-
+            'context.json')
         with open(target_file_path, 'w') as f:
-            f.write(yaml.dump(self))
+            json.dump(self.to_dict(), f, sort_keys=True, indent=4)
+            f.write('\n')
 
 
 def get_auth_header(username, password):
     header = {}
 
     if username and password:
-        credentials = '{0}:{1}'.format(username, password)
-        encoded_credentials = urlsafe_b64encode(credentials)
+        # encode/decode just to allow b64encode, which requires bytes
+        credentials = '{0}:{1}'.format(username, password).encode('utf-8')
+        encoded_credentials = b64encode(credentials).decode('utf-8')
         header = {
             constants.CLOUDIFY_AUTHENTICATION_HEADER:
                 constants.BASIC_AUTH_PREFIX + ' ' + encoded_credentials}
@@ -517,23 +537,24 @@ def get_auth_header(username, password):
 # Only the IP is required.
 # Note that not all attributes are allowed - username/password will be
 # the same for every node in the cluster.
-CLUSTER_NODE_ATTRS = ['manager_ip', 'rest_port', 'rest_protocol', 'ssh_port',
-                      'ssh_user', 'ssh_key']
+CLUSTER_NODE_ATTRS = ['host_ip', 'host_type', 'rest_port', 'rest_protocol',
+                      'ssh_port', 'ssh_user', 'ssh_key']
+_TRY_NEXT_NODE = object()
 
 
 class ClusterHTTPClient(HTTPClient):
-    default_timeout_sec = (5, None)
 
     def __init__(self, *args, **kwargs):
         profile = kwargs.pop('profile')
         super(ClusterHTTPClient, self).__init__(*args, **kwargs)
         if not profile.cluster:
             raise ValueError('Cluster client invoked for an empty cluster!')
-        self._cluster = list(profile.cluster)
+        self._cluster = list(profile.cluster.get(CloudifyNodeType.MANAGER))
         self._profile = profile
         first_node = self._cluster[0]
         self.cert = first_node.get('cert') or self.cert
         self.trust_all = first_node.get('trust_all') or self.trust_all
+        self.default_timeout_sec = self.default_timeout_sec or (5, None)
 
     def do_request(self, *args, **kwargs):
         # this request can be retried for each manager - if the data is
@@ -541,29 +562,55 @@ class ClusterHTTPClient(HTTPClient):
         copied_data = None
         if isinstance(kwargs.get('data'), types.GeneratorType):
             copied_data = itertools.tee(kwargs.pop('data'),
-                                        len(self._cluster))
+                                        len(self._cluster) + 1)
 
         if kwargs.get('timeout') is None:
             kwargs['timeout'] = self.default_timeout_sec
 
-        for node_index, node in list(enumerate(self._profile.cluster)):
+        if copied_data is not None:
+            kwargs['data'] = copied_data[-1]
+
+        manager_host = get_target_manager()
+        if manager_host:
+            self.host = manager_host
+            return super(ClusterHTTPClient, self).do_request(*args, **kwargs)
+
+        # First try with the main manager ip given when creating the profile
+        # with `cfy profiles use`
+        self.host = self._profile.manager_ip
+        response = self._try_do_request(*args, **kwargs)
+        if response is not _TRY_NEXT_NODE:
+            return response
+
+        for node_index, node in list(enumerate(
+                self._profile.cluster[CloudifyNodeType.MANAGER])):
+            if self._profile.manager_ip in [node['host_ip'], node['hostname']]:
+                continue
             self._use_node(node)
             if copied_data is not None:
                 kwargs['data'] = copied_data[node_index]
 
-            try:
-                return super(ClusterHTTPClient, self).do_request(*args,
-                                                                 **kwargs)
-            except (RemovedFromCluster, NotClusterMaster,
-                    requests.exceptions.ConnectionError):
+            response = self._try_do_request(*args, **kwargs)
+            if response is _TRY_NEXT_NODE:
                 continue
+            return response
 
-        raise CloudifyClientError('No active node in the cluster!')
+        raise CloudifyClientError('All cluster nodes are offline')
+
+    def _try_do_request(self, *args, **kwargs):
+        try:
+            return super(ClusterHTTPClient, self).do_request(*args,
+                                                             **kwargs)
+        except (requests.exceptions.ConnectionError,
+                CloudifyClientError) as e:
+            if isinstance(e, CloudifyClientError) and e.status_code != 502:
+                raise
+        return _TRY_NEXT_NODE
 
     def _use_node(self, node):
-        if node['manager_ip'] == self.host:
+        if node['host_ip'] == self.host:
             return
-        self.host = node['manager_ip']
+        self.host = node['host_ip']
         for attr in ['rest_port', 'rest_protocol', 'trust_all', 'cert']:
             new_value = node.get(attr)
             if new_value:
@@ -571,14 +618,16 @@ class ClusterHTTPClient(HTTPClient):
         self._update_profile(node)
 
     def _update_profile(self, node):
-        """Put the node at the start of the cluster list in profile.
+        """
+        Put the node at the start of the cluster list in profile.
 
         The client tries nodes in the order of the cluster list, so putting
         the node first will make the client try it first next time. This makes
-        the client always try the last-known-master first.
+        the client always try the last-known-active-manager first.
         """
-        self._profile.cluster.remove(node)
-        self._profile.cluster = [node] + self._profile.cluster
+        self._profile.cluster[CloudifyNodeType.MANAGER].remove(node)
+        self._profile.cluster[CloudifyNodeType.MANAGER] = (
+            [node] + self._profile.cluster[CloudifyNodeType.MANAGER])
         for node_attr in CLUSTER_NODE_ATTRS:
             if node_attr in node:
                 setattr(self._profile, node_attr, node[node_attr])
@@ -586,13 +635,14 @@ class ClusterHTTPClient(HTTPClient):
 
 
 class CloudifyClusterClient(CloudifyClient):
-    """A CloudifyClient that will retry the queries with the current master.
+    """
+    A CloudifyClient that will retry the queries with the current manager.
 
-    When a request fails with a connection error, or with a "not cluster
-    master" error, this will keep trying with every node in the cluster,
-    until it finds the cluster master.
+    When a request fails with a connection error, this will keep trying with
+    every node in the cluster, until it finds an active manager.
 
-    When the master is found, the profile will be updated with its address.
+    When an active manager is found, the profile will be updated with its
+    address.
     """
     def __init__(self, profile, *args, **kwargs):
         self._profile = profile
@@ -603,13 +653,56 @@ class CloudifyClusterClient(CloudifyClient):
         return ClusterHTTPClient(*args, **kwargs)
 
 
-def build_fabric_env(manager_ip, ssh_user, ssh_port, ssh_key_path):
-    return {
-        "host_string": manager_ip,
-        "user": ssh_user,
-        "port": ssh_port,
-        "key_filename": ssh_key_path
-    }
+@contextmanager
+def ssh_connection(host=None, user=None, key=None):
+    if Connection is None:
+        raise CloudifyCliError(
+            "SSH not available - fabric not installed")
+    if host is None:
+        host = profile.manager_ip
+    if user is None:
+        user = profile.ssh_user
+    if key is None:
+        key = profile.ssh_key
+    connect_kwargs = {}
+    if key:
+        connect_kwargs['key_filename'] = [key]
+    conn = Connection(
+        host=host,
+        user=user,
+        port=profile.ssh_port or 22,
+        connect_kwargs=connect_kwargs or None
+    )
+    try:
+        conn.open()
+        yield conn
+    except AuthenticationException as e:
+        if user:
+            user_message = user
+        else:
+            user_message = '{0} (from ssh config)'.format(conn.user)
+
+        if key:
+            key_message = key
+        elif conn.ssh_config.get('identityfile'):
+            key_message = '{0} (from ssh config)'.format(
+                ', '.join(conn.ssh_config['identityfile']))
+        else:
+            key_message = 'default'
+
+        raise CloudifyCliError(
+            "SSH: could not connect to {host} "
+            "(username: {user}, key: {key}): {exc}"
+            .format(
+                host=conn.host,
+                user=user_message,
+                key=key_message,
+                exc=e
+            )
+        )
+    finally:
+        conn.close()
 
 
 profile = get_profile_context(suppress_error=True)
+target_manager = None
