@@ -32,6 +32,7 @@ from .commands import config
 from .commands import cluster
 from .commands import install
 from .commands import plugins
+from .commands import filters
 from .commands import tenants
 from .commands import secrets
 from .commands import license
@@ -41,6 +42,7 @@ from .commands import uninstall
 from .commands import workflows
 from .commands import blueprints
 from .commands import executions
+from .commands import permissions
 from .commands import user_groups
 from .commands import deployments
 from .commands import certificates
@@ -106,6 +108,7 @@ def _register_commands():
     _cfy.add_command(license.license)
     _cfy.add_command(sites.sites)
     _cfy.add_command(certificates.certificates)
+    _cfy.add_command(filters.filters)
 
     deployments.deployments.add_command(deployments.manager_create)
     deployments.deployments.add_command(deployments.manager_delete)
@@ -115,6 +118,7 @@ def _register_commands():
     deployments.deployments.add_command(deployments.manager_get_update)
     deployments.deployments.add_command(deployments.manager_set_visibility)
     deployments.deployments.add_command(deployments.manager_set_site)
+    deployments.deployments.add_command(deployments.schedule)
 
     executions.executions.add_command(executions.manager_cancel)
 
@@ -125,6 +129,7 @@ def _register_commands():
         _cfy.add_command(install.manager)
         _cfy.add_command(uninstall.manager)
         _cfy.add_command(node_instances.node_instances)
+        _cfy.add_command(permissions.permissions)
 
         deployments.deployments.add_command(deployments.manager_inputs)
         deployments.deployments.add_command(deployments.manager_outputs)
